import numpy as np
import matplotlib.pyplot as plt
from typing import Union, List
from pysimenv.core.base import StaticObject
from pysimenv.core.system import DynSystem
from pysimenv.core.simulator import Simulator
from pysimenv.common.model import FlatEarthEnv
from pysimenv.common import orientation


class MulticopterDynamic(DynSystem):
    e3 = np.array([0., 0., 1.], dtype=np.float32)

    def __init__(self, initial_states: Union[list, tuple], m: float, J: np.ndarray):
        """
        :param initial_states: [p, v, R, omega] where
            p: position, (3,) numpy array
            v: velocity, (3,) numpy array
            R: rotation matrix, (3, 3) numpy array
            omega: angular velocity, (3,) numpy array
        :param m: mass, float
        :param J: inertia of mass, (3, 3) numpy array
        """
        super(MulticopterDynamic, self).__init__(
            initial_states={
                'p': initial_states[0], 'v': initial_states[1],
                'R': initial_states[2], 'omega': initial_states[3]}
        )
        self.m = m
        self.J = J
        self.grav_accel = FlatEarthEnv.grav_accel*self.e3

        self.state_vars['R'].attach_correction_fun(orientation.correct_orthogonality)

    # implement
    def _deriv(self, p, v, R, omega, u):
        """
        :param p: position, (3,) numpy array
        :param v: velocity, (3,) numpy array
        :param R: rotation matrix, (3, 3) numpy array
        :param omega: angular velocity, (3,) numpy array
        :param u: control input u = [f, tau], (4,) numpy array
        :return: [p_dot, v_dot, R_dot, omega_dot]
        """
        f = u[0]  # Total thrust
        tau = u[1:4]  # Moments

        p_dot = v
        v_dot = self.grav_accel - 1./self.m*(f*np.dot(R, self.e3))
        R_dot = np.matmul(R, self.hat(omega))
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, np.dot(self.J, omega)) + tau)

        return {'p': p_dot, 'v': v_dot, 'R': R_dot, 'omega': omega_dot}

    @classmethod
    def hat(cls, v):
        return np.array([
            [0, -v[2], v[1]],
            [v[2], 0, -v[0]],
            [-v[1], v[0], 0]
        ])

    @property
    def pos(self) -> np.ndarray:
        # position in the inertial frame
        return self.state('p').copy()

    @property
    def vel(self) -> np.ndarray:
        # velocity in the inertial frame
        return self.state('v').copy()

    @property
    def rotation(self) -> np.ndarray:
        # rotation matrix from the vehicle frame to the inertial frame R_iv
        R_iv = self.state('R').copy()
        return R_iv

    @property
    def quaternion(self) -> np.ndarray:
        R_iv = self.state('R')
        q = orientation.rotation_to_quaternion(np.transpose(R_iv))
        return q

    @property
    def euler_ang(self) -> np.ndarray:
        R_iv = self.state('R')
        eta = np.array(
            orientation.rotation_to_euler_angles(np.transpose(R_iv))
        )
        return eta

    @property
    def ang_vel(self) -> np.ndarray:
        # angular velocity of the vehicle frame with respect to the inertial frame
        return self.state('omega')

    def plot_euler_angles(self):
        time_list = self.history('t')
        rotation_list = self.history('R')

        data_num = rotation_list.shape[0]
        euler_angle_list = np.zeros((data_num, 3))

        for i in range(data_num):
            R_bi = np.transpose(rotation_list[i, :, :])
            euler_angle_list[i, :] = orientation.rotation_to_euler_angles(R_bi)

        fig = plt.figure()
        names = ['phi', 'theta', 'psi']
        for i in range(3):
            plt.subplot(3, 1, i + 1)
            plt.plot(time_list, np.rad2deg(euler_angle_list[:, i]), label="Actual")
            plt.xlabel("Time (s)")
            plt.ylabel(names[i])
            plt.grid()
            plt.legend()
        plt.suptitle("Euler angles")

        plt.draw()
        plt.pause(0.01)
        return fig


class QuadXThrustModel(object):
    def __init__(self, d_phi: float, d_theta: float, c_tau_f: float):
        self.d_phi = d_phi
        self.d_theta = d_theta
        self.c_tau_f = c_tau_f
        self.R_u = np.array([
            [1., 1., 1., 1.],
            [-d_phi/2., d_phi/2., d_phi/2., -d_phi/2.],
            [d_theta/2., -d_theta/2., d_theta/2., -d_theta/2.],
            [c_tau_f, c_tau_f, -c_tau_f, -c_tau_f]
        ])  # mapping matrix

    def convert(self, f_s: np.ndarray) -> np.ndarray:
        u = self.R_u.dot(f_s)
        return u


class QuadXMixer(object):
    def __init__(self, d_phi: float, d_theta: float, c_tau_f: float):
        self.d_phi = d_phi
        self.d_theta = d_theta
        self.c_tau_f = c_tau_f
        self.R_u = np.array([
            [1., 1., 1., 1.],
            [-d_phi/2., d_phi/2., d_phi/2., -d_phi/2.],
            [d_theta/2., -d_theta/2., d_theta/2., -d_theta/2.],
            [c_tau_f, c_tau_f, -c_tau_f, -c_tau_f]
        ])  # mapping matrix
        self.R_u_inv = np.linalg.inv(self.R_u)

    def convert(self, u_d: np.ndarray) -> np.ndarray:
        f_s = self.R_u_inv.dot(u_d)
        return f_s


class ActuatorFault(StaticObject):
    def __init__(self, t_list: List[float], alp_list: List[np.ndarray], rho_list: List[np.ndarray],
                 interval: Union[int, float] = -1):
        """
        :param t_list: [t_0, t_1, ..., t_{k-1}] time of fault occurrence
        :param alp_list: [lam_0: (M,) array, ..., lam_{k-1}] gain fault, M: the number of motors
        :param rho_list: [rho_0: (M,) array, ..., rho_{k-1}] bias fault, M: the number of motors
        :return:
        """
        super(ActuatorFault, self).__init__(interval)
        assert len(t_list) == len(alp_list), "Array sizes doesn't match."
        assert len(t_list) == len(rho_list), "Array sizes doesn't match."
        self.t_list = t_list
        self.alp_list = alp_list
        self.rho_list = rho_list

        self.alp = alp_list[0]
        self.rho = rho_list[0]
        self.next_ind = 1

    def _forward(self, f_s: np.ndarray):
        if self.next_ind < len(self.t_list):
            if self.time >= self.t_list[self.next_ind]:
                self.alp = self.alp_list[self.next_ind]
                self.rho = self.rho_list[self.next_ind]
                self.next_ind += 1
        f_s_star = self.alp*f_s + self.rho
        return f_s_star


def main():
    print("== Test for QuadrotorDynModel ==")
    m = 4.34
    J = np.diag([0.0820, 0.0845, 0.1377])

    pos = np.array([0., 0., -1.])
    vel = np.array([1., 0., 0.])
    R = np.identity(3)
    omega = np.array([0., 0., 0.1])
    initial_states = (pos, vel, R, omega)

    quadrotor = MulticopterDynamic(initial_states, m, J)

    u = np.array([45., 0., 0., 0.])

    simulator = Simulator(quadrotor)
    simulator.propagate(dt=0.01, time=10., save_history=True, u=u)

    quadrotor.default_plot()
    quadrotor.plot_euler_angles()
    plt.show()


if __name__ == "__main__":
    main()




import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from pysimenv.core.system import MultipleSystem
from pysimenv.missile.model import PlanarManVehicle2dof
from pysimenv.missile.guidance import Guidance2dim
from pysimenv.missile.util import RelKin2dim, CloseDistCond, closest_instant, lin_interp


class Engagement2dim(MultipleSystem):
    INTERCEPTED = 1
    MISSILE_STOP = 2
    IS_OUT_OF_VIEW = 3

    def __init__(self, missile: PlanarManVehicle2dof, target: PlanarManVehicle2dof, guidance: Guidance2dim):
        super(Engagement2dim, self).__init__()
        self.missile = missile
        self.target = target
        self.guidance = guidance
        self.rel_kin = RelKin2dim()
        self.close_dist_cond = CloseDistCond(r_threshold=10.0)

        self.attach_sim_objects([self.missile, self.target, self.guidance])

    # override
    def reset(self):
        super(Engagement2dim, self).reset()
        self.close_dist_cond.reset()

    # implement
    def initialize(self):
        x_M = self.missile.state('x')
        x_T = self.target.state('x')
        self.rel_kin.evaluate(x_M, x_T)

    # implement
    def _forward(self):
        x_M = self.missile.state('x')
        x_T = self.target.state('x')
        self.rel_kin.evaluate(x_M, x_T)
        self.close_dist_cond.evaluate(self.rel_kin.r)

        a_y_cmd = self.guidance.forward(self.missile, self.target, self.rel_kin)
        self.missile.forward(a_M_cmd=np.array([0., a_y_cmd]))
        self.target.forward()

        self._logger.append(
            t=self.time, r=self.rel_kin.r, sigma=self.rel_kin.sigma, lam=self.rel_kin.lam, omega=self.rel_kin.omega
        )

    # implement
    def check_stop_condition(self) -> Tuple[bool, int]:
        to_stop = False

        missile_stop, _ = self.missile.check_stop_condition()
        if self.intercepted():  # probable interception
            to_stop = True
            self.flag = self.INTERCEPTED

        if missile_stop:  # stop due to the missile
            to_stop = True
            self.flag = self.MISSILE_STOP

        return to_stop, self.flag

    def intercepted(self) -> bool:
        return self.close_dist_cond.check()

    def state_on_closest_instant(self):
        t = self.missile.history('t')
        x_M = self.missile.history('x')
        x_T = self.target.history('x')

        p_M = x_M[:, 0:2]
        p_T = x_T[:, 0:2]
        index_close, xi_close = closest_instant(p_M, p_T)

        t_close = lin_interp(t[index_close], t[index_close + 1], xi_close)
        x_M_close = lin_interp(x_M[index_close], x_M[index_close + 1], xi_close)
        x_T_close = lin_interp(x_T[index_close], x_T[index_close + 1], xi_close)

        return t_close, x_M_close, x_T_close

    def miss_distance(self) -> float:
        _, x_M_c, x_T_c = self.state_on_closest_instant()

        p_M_c = x_M_c[0:2]
        p_T_c = x_T_c[0:2]

        d_miss = np.linalg.norm(p_M_c - p_T_c)
        return d_miss

    def impact_angle(self) -> float:
        _, x_M_c, x_T_c = self.state_on_closest_instant()
        gamma_M_c = x_M_c[3]
        gamma_T_c = x_T_c[3]
        return gamma_M_c - gamma_T_c

    def impact_time(self) -> float:
        t_c, _, _ = self.state_on_closest_instant()
        return t_c

    def report(self):
        self.missile.report()
        if self.flag == self.INTERCEPTED:
            print("[engagement] The target has been intercepted!")
        else:
            print("[engagement] The target has been missed!")

        _, x_M_c, x_T_c = self.state_on_closest_instant()

        print("[engagement] Missile state on the closest instant: {:.2f}(m), {:.2f}(m), {:.2f}(m/s), {:.2f}(deg) ".
              format(x_M_c[0], x_M_c[1], x_M_c[2], np.rad2deg(x_M_c[3]))
              )
        print("[engagement] Target state on the closest instant: {:.2f}(m), {:.2f}(m), {:.2f}(m/s), {:.2f}(deg) \n".
              format(x_T_c[0], x_T_c[1], x_T_c[2], np.rad2deg(x_T_c[3]))
              )

    def report_miss_distance(self):
        d_miss = self.miss_distance()
        print("[engagement] Miss distance: {:.6f} (m)".format(d_miss))

    def report_impact_angle(self):
        gamma_imp = self.impact_angle()
        print("[engagement] Impact angle: {:.2f} (deg)".format(np.rad2deg(gamma_imp)))

    def report_impact_time(self):
        t_imp = self.impact_time()
        print("[engagement] Impact time: {:.2f} (s)".format(t_imp))

    def plot_path(self):
        fig_axs = dict()

        fig_ax = self.missile.plot_path()
        self.target.plot_path(fig_ax)

        fig1, ax1 = fig_ax['fig'], fig_ax['ax']
        fig1.suptitle("2-dim flight path")
        fig_axs['path'] = {'fig': fig1, 'ax': ax1}

        return fig_axs

    def plot_zem(self):
        fig_axs = dict()

        relKin = RelKin2dim()
        t = self.missile.history('t')
        x_M = self.missile.history('x')
        x_T = self.target.history('x')

        zem = []
        for i in range(x_M.shape[0]):
            x_M_ = x_M[i, :]
            x_T_ = x_T[i, :]
            relKin.evaluate(x_M_, x_T_)

            zem.append(relKin.zem)
        zem = np.array(zem)

        fig, ax = plt.subplots()
        ax.set_title("ZEM")
        ax.plot(t[:-1], zem[:-1], label="ZEM")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("ZEM (m)")
        ax.grid()
        ax.legend()

        fig_axs['Rel. Kin. add.'] = {'fig': fig, 'ax': ax}
        return fig_axs

    def plot_rel_kin(self):
        fig_axs = dict()

        t = self.history('t')
        r = self.history('r')
        sigma = self.history('sigma')
        lam = self.history('lam')
        omega = self.history('omega')
        
        fig, ax = plt.subplots(4, 1, figsize=(6, 8))
        ax[0].set_title("Rel. dist")
        ax[0].plot(t[:-1], r[:-1], label="Rel. dist")
        ax[0].set_xlabel("Time (s)")
        ax[0].set_ylabel("r (m)")
        ax[0].grid()

        ax[1].set_title("Look angle")
        ax[1].plot(t[:-1], np.rad2deg(sigma[:-1]), label="look angle")
        ax[1].set_xlabel("Time (s)")
        ax[1].set_ylabel("sigma (deg)")
        ax[1].grid()

        ax[2].set_title("LOS angle")
        ax[2].plot(t[:-1], np.rad2deg(lam[:-1]), label="LOS angle")
        ax[2].set_xlabel("Time (s)")
        ax[2].set_ylabel("lambda (deg)")
        ax[2].grid()

        ax[3].set_title("LOS rate")
        ax[3].plot(t[:-1], np.rad2deg(omega[:-1]), label="LOS rate")
        ax[3].set_xlabel("Time (s)")
        ax[3].set_ylabel("omega (deg/s)")
        ax[3].grid()
        fig.tight_layout()
        fig_axs['Rel. Kin.'] = {'fig': fig, 'ax': ax}

        return fig_axs





import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile2dof, PlanarNonManVehicle2dof
from pysimenv.missile.guidance import PurePNG2dim
from pysimenv.missile.engagement import Engagement2dim


def main():
    print("== Two dimensional pure PNG engagement for a stationary target ==")
    missile = PlanarMissile2dof(
        initial_state=[-5000., 3000., 300., np.deg2rad(-15.)]
    )
    target = PlanarNonManVehicle2dof(
        initial_state=[0., 0., 20., 0.]
    )
    missile.name = "missile"
    target.name = "target"

    # Method 1
    guidance = PurePNG2dim(N=3.0)
    model = Engagement2dim(missile, target, guidance)
    simulator = Simulator(model)
    simulator.propagate(dt=0.01, time=30., save_history=True)

    model.plot_path()
    model.plot_rel_kin()
    model.report()
    model.report_miss_distance()
    model.report_impact_angle()
    model.report_impact_time()
    model.save_log_file('./data/planar_png/')
    # model.load_log_file('./data/planar_png/')

    plt.show()


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from pysimenv.core.base import ArrayType
from pysimenv.core.system import DynSystem
from pysimenv.common.util import wrap_to_pi


class PlanarManVehicle2dof(DynSystem):
    """
    A 2-dof model for a maneuvering vehicle in the plane
    state x: [p_x, p_y, V, gamma]
    control u: [a_x, a_y]
    where [p_x, p_y] is the position expressed in the inertial frame and
    [a_x, a_y] is the acceleration expressed in the velocity frame
    """
    def __init__(self, initial_state: ArrayType):
        super(PlanarManVehicle2dof, self).__init__(initial_states={'x': initial_state})
        self.name = "planar_man_vehicle"

        def angle_correction_fun(state: np.ndarray) -> np.ndarray:
            state[3] = wrap_to_pi(state[3])  # gamma should be in [-pi, pi] rad
            return state
        self.state_vars['x'].attach_correction_fun(angle_correction_fun)

    def _deriv(self, x: np.ndarray, u: np.ndarray = np.zeros(2)):
        """
        :param x: [p_x, p_y, V, gamma]
        :param u: [a_x, a_y]
        :return: x_dot
        """
        V = x[2]
        gamma = x[3]

        a_x = u[0]
        a_y = u[1]

        p_x_dot = V*np.cos(gamma)
        p_y_dot = V*np.sin(gamma)
        V_dot = a_x

        if abs(V) < 1e-4:
            gamma_dot = 0
        else:
            gamma_dot = a_y/V

        return {'x': np.array([p_x_dot, p_y_dot, V_dot, gamma_dot])}

    @property
    def p(self) -> np.ndarray:
        """
        :return: [p_x, p_y]
        """
        return self.state('x')[0:2]

    @property
    def v(self) -> np.ndarray:
        """
        :return: [v_x, v_y]
        """
        x = self.state('x')
        V = x[2]
        gamma = x[3]

        return np.array([V*np.cos(gamma), V*np.sin(gamma)])

    @property
    def V(self) -> float:
        """
        :return: V
        """
        return self.state('x')[2]

    @property
    def gamma(self) -> float:
        """
        :return: gamma
        """
        return self.state('x')[3]

    def plot(self):
        var_keys = {'x', 'u'}
        var_ind_dict = {'x': [0, 1, 2, 3], 'u': [0, 1]}
        var_names_dict = {
            'x': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)'],
            'u': ['a_x (m/s**2)', 'a_y (m/s**2)']
        }
        fig_axs = self.default_plot(var_keys=var_keys, var_ind_dict=var_ind_dict, var_names_dict=var_names_dict)
        return fig_axs

    def plot_path(self, fig_ax=None, show=False):
        if fig_ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("p_x (m)")
            ax.set_ylabel("p_y (m)")
            ax.grid()
        else:
            fig = fig_ax['fig']
            ax = fig_ax['ax']

        x = self.history('x')
        p = x[:, 0:2]

        ax.plot(p[:, 0], p[:, 1], label=self.name + " path")
        ax.set_aspect('equal')
        ax.legend()
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return {'fig': fig, 'ax': ax}


class PlanarNonManVehicle2dof(PlanarManVehicle2dof):
    def __init__(self, initial_state: ArrayType):
        super(PlanarNonManVehicle2dof, self).__init__(initial_state)
        self.name = "planar_non_man_vehicle"

    # override
    def plot(self):
        var_keys = {'x'}
        var_ind_dict = {'x': [0, 1, 2, 3]}
        var_names_dict = {
            'x': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)']
        }
        fig_axs = self.default_plot(var_keys, var_ind_dict, var_names_dict)
        return fig_axs

    # override
    def plot_path(self, fig_ax=None, show=False):
        if fig_ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("p_x (m)")
            ax.set_ylabel("p_y (m)")
            ax.grid()
        else:
            fig = fig_ax['fig']
            ax = fig_ax['ax']

        x = self.history('x')
        p = x[:, 0:2]

        if abs(self.V) < 1e-4:
            ax.plot(p[0, 0], p[0, 1], marker='o', label=self.name)
        else:
            ax.plot(p[:, 0], p[:, 1], label=self.name + " path")
        ax.set_aspect('equal')
        ax.legend()
        fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)
        return {'fig': fig, 'ax': ax}


class PlanarMissile2dof(PlanarManVehicle2dof):
    NORMAL = 0
    STALLED = 1
    COLLIDED = 2

    def __init__(self, initial_state: ArrayType):
        super(PlanarMissile2dof, self).__init__(initial_state)
        self.name = "missile"
        self.fov_limit = np.inf
        self.acc_limit = np.array([
            [-np.inf, -np.inf],
            [np.inf, np.inf]
        ])
        self.ground_elevation = -np.inf

    def _forward(self, a_M_cmd: np.ndarray = np.zeros(2)):
        a_M = np.clip(a_M_cmd, self.acc_limit[0], self.acc_limit[1])
        super(PlanarMissile2dof, self)._forward(u=a_M)
        self._logger.append(a_M_cmd=a_M_cmd)

    # implement
    def check_stop_condition(self) -> Tuple[bool, int]:
        to_stop = False
        if self.is_stalled():
            to_stop = True
            self.flag = self.STALLED
        if self.is_collided():
            to_stop = True
            self.flag = self.COLLIDED
        return to_stop, self.flag

    def is_stalled(self) -> bool:
        return self.V < 10  # when the speed is less than 10m/s

    def is_collided(self) -> bool:
        return self.p[1] < self.ground_elevation - 0.5

    def is_out_of_view(self, sigma: float) -> bool:
        return abs(sigma) > self.fov_limit

    def report(self):
        np.set_printoptions(precision=2, suppress=True)
        x_M_f = self.state('x')

        print("[{:s}] Final state: {:.2f}(m), {:.2f}(m), {:.2f}(m/s), {:.2f}(deg)".format(
            self.name, x_M_f[0], x_M_f[1], x_M_f[2], np.rad2deg(x_M_f[3])))

        if self.flag == self.NORMAL:
            print("[{:s}] Status: normal \n".format(self.name))
        elif self.flag == self.STALLED:
            print("[{:s}] Status: stalled \n".format(self.name))
        elif self.flag == self.COLLIDED:
            print("[{:s}] Status: collided \n".format(self.name))



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarManVehicle2dof, PlanarNonManVehicle2dof


def main():
    print("== Test for PlanarManVehicle2dof ==")
    man_vehicle = PlanarManVehicle2dof(
        initial_state=[-5000., 0., 200., np.deg2rad(30.)])
    simulator = Simulator(man_vehicle)
    simulator.propagate(dt=0.01, time=30., save_history=True, u=np.array([0., -5.]))
    man_vehicle.default_plot()
    man_vehicle.plot_path(show=True)

    print("== Test for PlanarNonManVehicle2dof ==")
    non_man_vehicle = PlanarNonManVehicle2dof(
        initial_state=[0., 0., 10., np.deg2rad(175.)]
    )
    simulator = Simulator(non_man_vehicle)
    simulator.propagate(dt=0.01, time=30., save_history=True)
    non_man_vehicle.default_plot()
    non_man_vehicle.plot_path(show=True)


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib.pyplot as plt
from typing import Union
from pysimenv.core.base import SimObject, StaticObject
from pysimenv.core.system import DynObject
from pysimenv.missile.model import PlanarManVehicle2dof
from pysimenv.missile.util import RelKin2dim


class Guidance2dim(SimObject):
    def __init__(self, interval: Union[int, float] = -1):
        super(Guidance2dim, self).__init__(interval=interval)

    # to be implemented
    def _forward(self, missile: PlanarManVehicle2dof, target: PlanarManVehicle2dof, rel_kin: RelKin2dim) -> float:
        raise NotImplementedError


class PurePNG2dim(StaticObject, Guidance2dim):
    def __init__(self, N: float = 3.0, interval: Union[int, float] = -1):
        super(PurePNG2dim, self).__init__(interval=interval)
        self.N = N

    # implement
    def _forward(self, missile: PlanarManVehicle2dof, target: PlanarManVehicle2dof, rel_kin: RelKin2dim) -> float:
        """
        :return: acceleration command a_y_cmd
        """
        V_M = missile.V  # speed of the missile
        omega = rel_kin.omega  # LOS rate

        a_y_cmd = self.N*V_M*omega
        return a_y_cmd


class IACBPNG(DynObject, Guidance2dim):
    """
    Biased PNG with terminal-angle constraint (impact angle control)
    """
    def __init__(self, N: float, tau: float, theta_M_0: float, theta_M_f: float, theta_T: float, lam_0: float,
                 sigma_max: float = float('inf'), b_max: float = float('inf')):
        super(IACBPNG, self).__init__(initial_states={'B': np.array([0.])})

        # Guidance parameters
        self.N = N
        self.tau = tau
        self.theta_M_0 = theta_M_0
        self.theta_M_f = theta_M_f
        self.theta_T = theta_T
        self.lam_0 = lam_0
        self.sigma_max = sigma_max
        self.b_max = b_max

        self.phase = 0  # 0: BPNG, 1: DPP, 2: BPNG

    @classmethod
    def reference_value(cls, theta_M_0, theta_M_f, theta_T, lam_0, N, V_M, V_T):
        theta_M_0_bar = theta_M_0 - lam_0
        theta_M_f_bar = theta_M_f - lam_0
        theta_T_bar = theta_T - lam_0

        lam_f_bar = np.arctan(
            (V_M*np.sin(theta_M_f_bar) - V_T*np.sin(theta_T_bar))/(V_M*np.cos(theta_M_f_bar) - V_T*np.cos(theta_T_bar))
        )
        B_ref = theta_M_f_bar - theta_M_0_bar - N*lam_f_bar
        return B_ref

    def _forward(self, missile: PlanarManVehicle2dof, target: PlanarManVehicle2dof, rel_kin: RelKin2dim) -> float:
        V_M = missile.V
        V_T = target.V
        omega = rel_kin.omega
        sigma = rel_kin.sigma

        B_ref = IACBPNG.reference_value(
            self.theta_M_0, self.theta_M_f, self.theta_T, self.lam_0, self.N, V_M, V_T)
        B = self.state('B')[0]

        b_1 = 1./self.tau*(B_ref - B)
        b_2 = (1. - self.N)*omega
        if self.phase == 0:
            # BPNG
            b = b_1
            if np.abs(sigma) > self.sigma_max - np.deg2rad(0.01):
                self.phase = 1
        elif self.phase == 1:
            # DPP
            b = b_2
            if np.abs(b_1) < np.abs(b_2):
                self.phase = 2
        else:
            # BPNG
            b = b_1

        b = np.clip(b, -self.b_max, self.b_max)
        a_png = self.N*V_M*omega
        a_b = V_M*b
        a_M = a_png + a_b

        self.state_vars['B'].set_deriv(deriv=b)
        self._logger.append(t=self.time, B_ref=B_ref, B=B, b=b, a_png=a_png, a_b=a_b)
        return a_M

    def plot_bias(self):
        t = self.history('t')
        b = self.history('b')

        fig, ax = plt.subplots()
        ax.plot(t, b)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Bias, b (rad/s)")
        ax.grid()

    def plot_bias_integral(self):
        t = self.history('t')
        B_ref = self.history('B_ref')
        B = self.history('B')

        fig, ax = plt.subplots()
        ax.plot(t, B_ref, linestyle='--')
        ax.plot(t, B)
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Integral Value of Bias, B (rad)")
        ax.grid()



"""
Description for Package
"""
from pysimenv.multicopter.model import MulticopterDynamic, QuadXThrustModel, QuadXMixer, ActuatorFault
from pysimenv.multicopter.control import FLVelControl, QuaternionAttControl, QuaternionPosControl
from pysimenv.multicopter.estimator import FixedTimeFaultEstimator

__all__ = ['multicopter']



import matplotlib.pyplot as plt
import numpy as np
from typing import Union, List, Tuple, Optional
from pysimenv.core.base import SimObject, StaticObject, ArrayType
from pysimenv.core.system import DynObject, MultipleSystem, DynSystem
from pysimenv.common import orientation


class SignalGenerator(StaticObject):
    def __init__(self, shaping_fun, interval: Union[int, float] = -1):
        super(SignalGenerator, self).__init__(interval=interval)
        self.shaping_fun = shaping_fun

    # implementation
    def _forward(self) -> np.ndarray:
        return self.shaping_fun(self.time)


class Scope(StaticObject):
    def __init__(self):
        super(Scope, self).__init__()

    # implementation
    def _forward(self, **kwargs):
        self._logger.append(t=self.time, **kwargs)

    def plot(self, show: bool = False):
        var_names = list(self._logger.keys())
        var_names.remove('t')

        t = self.history('t')
        for var_name in var_names:
            var = self.history(var_name)

            fig, ax = plt.subplots()
            for i in range(var.shape[1]):
                ax.plot(t, var[:, i], label=var_name + "_" + str(i))
                ax.set_xlabel("Time (s)")
                ax.set_ylabel("Value")
                ax.grid()
                ax.legend()
            fig.tight_layout()

        if show:
            plt.show()
        else:
            plt.draw()
            plt.pause(0.01)


class FeedbackControl(MultipleSystem):
    def __init__(self, system: DynObject, control: StaticObject):
        super(FeedbackControl, self).__init__()
        self.system = system
        self.control = control

        self.attach_sim_objects([system, control])

    # implement
    def _forward(self, r: Optional[np.ndarray] = None):
        y = self.system.output
        if r is None:
            u_fb = self.control.forward(y)
        else:
            u_fb = self.control.forward(y, r)
        self.system.forward(u=u_fb)


class Sequential(MultipleSystem):
    def __init__(self, obj_list: Union[List[SimObject], Tuple[SimObject]]):
        super(Sequential, self).__init__()
        assert len(obj_list) > 0, "(sequential) Invalid obj_list!"
        self.obj_list = list(obj_list)
        self.first_obj = obj_list[0]
        self.other_obj_list = obj_list[1:]

        self.attach_sim_objects(obj_list)

    # implement
    def _forward(self, **kwargs):
        self.first_obj.forward(**kwargs)
        out = self.first_obj.output
        for obj in self.other_obj_list:
            if out is None:
                obj.forward()
                out = obj.output
            elif isinstance(out, dict):
                obj.forward(**out)
                out = obj.output
            else:
                obj.forward(u=out)
                out = obj.output
        return out


class FlatEarthEnv(object):
    grav_accel = 9.805


class LinSys(DynSystem):
    def __init__(self, x_0: ArrayType, A: np.ndarray, B: Optional[np.ndarray] = None, C: Optional[np.ndarray] = None):
        super(LinSys, self).__init__(initial_states={'x': x_0})
        self.A = A
        self.B = B
        if C is None:
            C = np.eye(self.state_dim, dtype=np.float32)
        self.C = C

    # implement
    def _deriv(self, x: np.ndarray, u: Union[None, np.ndarray] = None):
        if self.B is None or u is None:
            x_dot = self.A.dot(x)
        else:
            x_dot = self.A.dot(x) + self.B.dot(u)

        return {'x': x_dot}

    # implement
    def _output(self) -> np.ndarray:
        return self.C.dot(self.state('x'))


class FirstOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, tau: float):
        A = np.array([[-1./tau]])
        B = np.array([[1./tau]])

        super(FirstOrderLinSys, self).__init__(x_0, A, B)
        self.tau = tau

    # implement
    def _output(self) -> Union[None, tuple, np.ndarray]:
        return self.state('x')[0]


class SecondOrderLinSys(LinSys):
    def __init__(self, x_0: ArrayType, zeta: float, omega: float):
        A = np.array([
            [0., 1.],
            [-omega**2, -2*zeta*omega]
        ], dtype=np.float32)
        B = np.array([[0.], [omega**2]], dtype=np.float32)

        super(SecondOrderLinSys, self).__init__(x_0, A, B)
        self.zeta = zeta
        self.omega = omega

    # implement
    def _output(self) -> Union[None, tuple, np.ndarray]:
        return self.state('x')[0]


class Integrator(DynSystem):
    def __init__(self, initial_state: Union[ArrayType]):
        def deriv_fun(x: np.ndarray, u: Union[float, np.ndarray]):
            if isinstance(u, float):
                u = np.array([u])
            return {'x': u}

        super(Integrator, self).__init__({'x': initial_state}, deriv_fun)


class Differentiator(StaticObject):
    def __init__(self):
        super(Differentiator, self).__init__()
        self.u_prev = None
        self.t_prev = None

    # implement
    def _forward(self, u: np.ndarray):
        if self.u_prev is None:
            deriv = np.zeros_like(u)
        else:
            deriv = (u - self.u_prev)/(self.time - self.t_prev)

        if self._sim_clock.major_time_step:
            self.u_prev = u.copy()
            self.t_prev = self.time

        return deriv


class PIDControl(DynObject):
    def __init__(self, k_p: Union[float, np.ndarray], k_i: Union[float, np.ndarray], k_d: Union[float, np.ndarray],
                 windup_limit=None,
                 interval: Union[int, float] = -1):
        super(PIDControl, self).__init__(initial_states={'e_i': np.zeros_like(k_p)}, interval=interval)
        self.k_p = k_p
        self.k_i = k_i
        self.k_d = k_d
        self.windup_limit = windup_limit
        if self.windup_limit:
            self.state_vars['e_i'].attach_correction_fun(self.clip_windup)

        self.e_prev = None
        self.t_prev = None

    # implement
    def _forward(self, e: np.ndarray) -> np.ndarray:
        self.state_vars['e_i'].set_deriv(deriv=e)

        e_i = self.state('e_i')
        e_d = np.zeros_like(e) if self.e_prev is None else (e - self.e_prev)/(self.time - self.t_prev)
        u_pid = self.k_p*e + self.k_i*e_i + self.k_d*e_d

        if self._sim_clock.major_time_step:
            self.e_prev = e.copy()
            self.t_prev = self.time

        return u_pid

    def clip_windup(self, e_i: np.ndarray) -> np.ndarray:
        return np.clip(e_i, self.windup_limit, -self.windup_limit)


class Dyn6DOF(DynSystem):
    def __init__(self, p_0: np.ndarray, v_b_0: np.ndarray, q_0: np.ndarray, omega_0: np.ndarray,
                 m: float, J: np.ndarray):
        super(Dyn6DOF, self).__init__(initial_states={'p': p_0, 'v_b': v_b_0, 'q': q_0, 'omega': omega_0})
        self.m = m
        self.J = J

        def normalize(v: np.ndarray) -> np.ndarray:
            v_norm = np.linalg.norm(v)
            if v_norm < 1e-6:
                return np.zeros_like(v)
            else:
                return v/v_norm

        self.state_vars['q'].attach_correction_fun(normalize)

    @property
    def R(self) -> np.ndarray:
        # rotation matrix representing the rotation from the body frame to the inertial frame
        q = self.state('q')
        eta, epsilon = q[0], q[1:4]
        S_epsilon = orientation.hat(epsilon)
        R_ib = np.identity(3) + 2*eta*S_epsilon + 2*np.matmul(S_epsilon, S_epsilon)
        return R_ib

    @property
    def v_i(self) -> np.ndarray:
        v_b = self.state('v_b')
        return self.R.dot(v_b)

    # implement
    def _deriv(self, p, v_b, q, omega, f_b: np.ndarray, m_b: np.ndarray):
        p_dot = self.v_i
        v_b_dot = -np.cross(omega, v_b) + 1./self.m

        Q = np.array([
            [-q[1], -q[2], -q[3]],
            [q[0], -q[3], q[2]],
            [q[3], q[0], -q[1]],
            [-q[2], q[1], q[0]]
        ])
        q_dot = 0.5*Q.dot(omega)
        omega_dot = np.linalg.solve(self.J, -np.cross(omega, np.dot(self.J, omega)) + m_b)
        return {'p': p_dot, 'v_b': v_b_dot, 'q': q_dot, 'omega': omega_dot}



import numpy as np
from pysimenv.core.simulator import Simulator
from pysimenv.missile.model import PlanarMissile2dof


def main():
    print("== Test for PlanarMissile2dof ==")
    missile = PlanarMissile2dof(
        initial_state=[-5000., 0., 200., np.deg2rad(30.)])
    simulator = Simulator(missile)
    simulator.propagate(dt=0.01, time=50., save_history=True, a_M_cmd=np.array([0., -5.]))

    missile.report()
    missile.plot()
    missile.plot_path(show=True)


if __name__ == "__main__":
    main()




import numpy as np
import matplotlib.pyplot as plt
from typing import Union, Callable, Optional, Tuple
from pysimenv.core.util import SimClock, Timer, Logger
from pysimenv.core.error import NoSimClockError


ArrayType = Union[list, tuple, np.ndarray]


class StateVariable(object):
    def __init__(self, state: ArrayType):
        self.state = np.array(state)
        self.deriv = np.zeros_like(state)
        self._rk4_buffer = []
        self.correction_fun = None

    def apply_state(self, state: ArrayType):
        self.state = np.array(state)

    @property
    def size(self):
        return self.state.size

    def set_deriv(self, deriv: np.ndarray):
        self.deriv = deriv

    def rk4_update_1(self, dt: float):
        x_0 = np.copy(self.state)
        k_1 = np.copy(self.deriv)

        self._rk4_buffer.clear()
        self._rk4_buffer.append(x_0)
        self._rk4_buffer.append(k_1)
        self.state = x_0 + dt / 2 * k_1  # x = x0 + dt/2*k1
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_2(self, dt: float):
        x_0 = self._rk4_buffer[0]
        k_2 = np.copy(self.deriv)

        self._rk4_buffer.append(k_2)
        self.state = x_0 + dt / 2 * k_2  # x = x0 + dt/2*k2
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_3(self, dt: float):
        x_0 = self._rk4_buffer[0]
        k_3 = np.copy(self.deriv)

        self._rk4_buffer.append(k_3)
        self.state = x_0 + dt / 2 * k_3  # x = x0 + dt/2*k3
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_4(self, dt: float):
        x_0, k_1, k_2, k_3 = self._rk4_buffer[:]
        k_4 = np.copy(self.deriv)

        self.state = x_0 + dt * (k_1 + 2 * k_2 + 2 * k_3 + k_4) / 6
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def attach_correction_fun(self, correction_fun: Callable[[np.ndarray], np.ndarray]):
        self.correction_fun = correction_fun

    @staticmethod
    def test():
        state_var = StateVariable([1.])
        def deriv_fun(x): return -x
        dt = 0.01

        time_list = []
        sim_state_list = []
        t = 0
        for i in range(1000):
            time_list.append(t)
            sim_state_list.append(state_var.state)

            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_1(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_2(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_3(dt)
            state_var.set_deriv(deriv_fun(state_var.state))
            state_var.rk4_update_4(dt)
            t = t + dt

        time_list = np.array(time_list)
        sim_state_list = np.array(sim_state_list)
        true_state_list = np.exp(-time_list)

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.plot(time_list, sim_state_list[:, 0], color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("State")
        plt.grid()

        plt.subplot(2, 1, 2)
        plt.plot(time_list, sim_state_list[:, 0] - true_state_list, color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("Integration error")
        plt.grid()

        plt.subplots_adjust(bottom=0.1, left=0.1, top=0.9, hspace=0.5)
        plt.show()


class SimObject(object):
    FLAG_OPERATING = 0

    def __init__(self, interval: Union[int, float] = -1):
        self.flag: int = SimObject.FLAG_OPERATING
        self._sim_clock: Optional[SimClock] = None
        self._log_timer: Optional[Timer] = None
        self._timer = Timer(event_time_interval=interval)
        self._logger = Logger()
        self._last_output = None

    def attach_sim_clock(self, sim_clock: SimClock):
        self._sim_clock = sim_clock
        self._timer.attach_sim_clock(sim_clock)
        self._timer.turn_on()

    def attach_log_timer(self, log_timer: Timer):
        self._log_timer = log_timer

    def initialize(self):
        self._initialize()

    def _initialize(self):
        pass

    def detach_sim_clock(self):
        self._sim_clock = None
        self._timer.turn_off()
        self._timer.detach_sim_clock()

    def detach_log_timer(self):
        self._log_timer = None
    
    def reset(self):
        self.flag = SimObject.FLAG_OPERATING
        self._timer.reset()
        self._logger.clear()

    def check_sim_clock(self):
        assert self._sim_clock is not None, "Attach a sim_clock first!"

    def check_log_timer(self):
        assert self._log_timer is not None, "Attach a log_timer first!"

    @property
    def time(self) -> float:
        if self._sim_clock is None:
            raise NoSimClockError
        return self._sim_clock.time

    def forward(self, *args, **kwargs):
        self._timer.forward()
        output = self._forward(*args, **kwargs)

        if self._timer.is_event:
            self._last_output = output

        return self._last_output

    # to be implemented
    def _forward(self, *args, **kwargs):
        return NotImplementedError

    @property
    def output(self) -> Union[None, tuple, np.ndarray]:
        return self._last_output

    def history(self, *args):
        """
        :param args: variable names
        :return:
        """
        return self._logger.get(*args)

    # to be implemented
    def check_stop_condition(self) -> Tuple[bool, int]:
        to_stop = False
        flag = self.flag
        return to_stop, flag


class StaticObject(SimObject):
    def __init__(self, interval: Union[int, float] = -1, eval_fun=None):
        super(StaticObject, self).__init__(interval)
        self.eval_fun = eval_fun

    # override
    def forward(self, *args, **kwargs):
        self._timer.forward()
        if self._timer.is_event:
            self._last_output = self._forward(*args, **kwargs)

        return self._last_output

    # may be implemented
    def _forward(self, *args, **kwargs):
        if self.eval_fun is None:
            raise NotImplementedError
        else:
            return self.eval_fun(*args, **kwargs)


class BaseFunction(object):
    # to be implemented
    def evaluate(self, *args, **kwargs):
        raise NotImplementedError


if __name__ == "__main__":
    StateVariable.test()
